from doct import Document
import ujson
import time as ttime
from amostra.client import conf
from uuid import uuid4
from os.path import expanduser
import mongoquery
from .amutils import doc_or_uid_to_uid


def _find_local(fname, qparams, as_doct=False):
    """Find a document created using the local framework
    Parameters
    -----------
    fname: str
        Name of the query should be run
    qparams: dict
        Query parameters. Similar to online query methods

    Yields
    ------------
    c: doct.Document, StopIteration
        Result of the query if found

    """
    res_list = []
    try:
        with open(fname, 'r') as fp:
            local_payload = ujson.load(fp)
        qobj = mongoquery.Query(qparams)
        for i in local_payload:
            if qobj.match(i):
                res_list.append(i)
    except FileNotFoundError:
        raise RuntimeWarning('Local file {} does not exist'.format(fname))
    if as_doct:
        for c in res_list:
            yield Document(fname.split('.')[0], c)
    else:
        for c in res_list:
            yield c


def _update_local(fname, qparams, replacement):
    """Update a document created using the local framework
    Parameters
    -----------
    fname: str
        Name of the query should be run
    qparams: dict
        Query parameters. Similar to online query methods
    replacement: dict
        Fields/value pair to be updated. Beware of disallowed fields
        such as time and uid
    """
    try:
        with open(fname, 'r') as fp:
            local_payload = ujson.load(fp)
        qobj = mongoquery.Query(qparams)
        for _sample in local_payload:
            try:
                if qobj.match(_sample):
                    for k, v in replacement.items():
                        _sample[k] = v
            except mongoquery.QueryError:
                pass
        with open(fname, 'w') as fp:
            ujson.dump(local_payload, fp)
    except FileNotFoundError:
        raise RuntimeWarning('Local file {} does not exist'.format(fname))


class LocalSampleReference:
    """Handle sample information locally via json files"""
    def __init__(self, top_dir=conf.local_conn_config['top']):
        self.top_dir = top_dir

    @property
    def sample_list(self):
        try:
            with open(self._samp_fname, 'r') as fp:
                _sample_list = ujson.load(fp)
            return _sample_list
        except (FileNotFoundError, ValueError):
            return []

    def create(self, name=None, time=None, uid=None, container=None,
               **kwargs):
        """Create a sample locally

        Parameters
        ----------
        name: str
            Name of the sample
        time: float
            Timestamp generated by the client
        uid: str
            Unique identifier for this sample
        container: str, doct.Document
            The container/group sample is contained within

        Returns
        -------
        payload: dict
            Document dict that was inserted
        """
        # TODO: Allow container to be an object
        tmp =  self.sample_list
        if container:
            container = doc_or_uid_to_uid(container)
        payload = dict(uid=uid if uid else str(uuid4()),
                       name=name, time=time if time else ttime.time(),
                       container=container if container else 'NULL',
                       **kwargs)

        tmp.append(payload)
        with open(self._samp_fname, 'w+') as fp:
            ujson.dump(tmp, fp)
        return payload

    @property
    def _samp_fname(self):
        return expanduser(self.top_dir + '/samples.json')

    def update(self, query, replacement):
        _update_local(self._samp_fname, query, replacement)

    def find(self, **kwargs):
        return _find_local(self._samp_fname, kwargs)


class LocalRequestReference:
    """Handle request information locally via json files"""
    def __init__(self, top_dir=conf.local_conn_config['top']):
        """Try to access files. If no file exists, create empty list

        Parameters
        ----------
        top_dir: str
            Directory where local files are stored
        """
        self.top_dir = top_dir

    @property
    def request_list(self):
        try:
            with open(self._req_fname, 'r') as fp:
                _req_list = ujson.load(fp)
            return _req_list
        except (FileNotFoundError, ValueError):
            return []

    @property
    def _req_fname(self):
        return expanduser(self.top_dir + '/requests.json')

    def create(self, sample=None, time=None, uid=None, state='active',
               seq_num=0, **kwargs):
        """Create a request locally

        Parameters
        ----------
        sample: str, doct.Document, optional
            Sample this request points to
        time: float
            Timestamp generated by the client
        uid: str
            Unique identifier for this sample
        state: str
            Defines whether this request is active or not
        seq_num: int
            Shows the order of request in the list of requests

        Returns
        -------
        local_payload: dict
            Request dict that was inserted locally

        """
        tmp = self.request_list
        if sample:
            sample = doc_or_uid_to_uid(sample)
        payload = dict(uid=uid if uid else str(uuid4()),
                       sample=sample if sample else 'NULL',
                       time=time if time else ttime.time(), state=state,
                       seq_num=seq_num, **kwargs)
        tmp.append(payload)
        with open(self._req_fname, 'w+') as fp:
            ujson.dump(tmp, fp)
        return payload

    def update(self, query, replacement):
        _update_local(self._req_fname, query, replacement)

    def find(self, **kwargs):
        return _find_local(self._req_fname, kwargs)


class LocalContainerReference:
    """Handle container information locally via json files"""
    def __init__(self, top_dir=conf.local_conn_config['top']):
        self.top_dir = top_dir
    @property
    def container_list(self):
        try:
            with open(self._cont_fname, 'r') as fp:
                _cont_list = ujson.load(fp)
            return _cont_list
        except (FileNotFoundError, ValueError):
            return []

    @property
    def _cont_fname(self):
        return expanduser(self.top_dir + '/containers.json')

    def create(self, uid=None, time=None, container=None, **kwargs):
        """ Create a container locally.

        Parameters
        ----------
        time: float
            Timestamp generated by the client
        uid: str
            Unique identifier for this sample
        container: str, doct.Document, optional
            Container this container is contained within

        Returns
        -------
        payload: dict
            Document dict that was inserted
        """
        tmp = self.container_list
        if container:
            container = doc_or_uid_to_uid(container)
        payload = dict(uid=uid if uid else str(uuid4()),
                       container=container if container else 'NULL',
                       time=time if time else ttime.time(), **kwargs)
        tmp.append(payload)
        with open(self._cont_fname, 'w+') as fp:
            ujson.dump(tmp, fp)
        return payload

    def update(self, query, replacement):
        _update_local(self._cont_fname, query, replacement)

    def find(self, **kwargs):
        return _find_local(self._cont_fname, kwargs)



import uuid
import time as ttime
import pytest
import time
from ..testing import amostra_setup, amostra_teardown
from ..client.api import ContainerReference
from requests.exceptions import HTTPError, ConnectionError, RequestException
from ..testing import TESTING_CONFIG
from doct import Document

# TODO: Replace routines with a class for testing


def teardown():
    amostra_teardown()


def test_container_constructor():
    # attempt empty reference create
    s2 = ContainerReference()


def test_connection_switch():
    c = ContainerReference()
    c.host = 'caesar'
    pytest.raises(RequestException, c.create)
    c.host = TESTING_CONFIG['host']
    c.port = TESTING_CONFIG['port']
    c.create()


def test_container_create():
    c = ContainerReference(TESTING_CONFIG['host'], TESTING_CONFIG['port'])
    ast_cont = {'name': 'obelix', "dog": 'hidefix', 'time': ttime.time(),
                'container': 'gauls', 'uid': str(uuid.uuid4())}
    cont1 = c.create(**ast_cont)
    assert cont1 == ast_cont['uid']


def test_duplicate_container():
    c = ContainerReference(TESTING_CONFIG['host'], TESTING_CONFIG['port'])
    m_uid = str(uuid.uuid4())
    ast_cont = {'name': 'obelix', "dog": 'hidefix', 'time': ttime.time(),
                'container': 'gauls', 'uid': m_uid}
    cont1 = c.create(**ast_cont)
    pytest.raises(HTTPError, c.create, uid=cont1)
    ast_cont = {'name': 'obelix', "dog": 'hidefix', 'time': ttime.time(),
                'container': 'gauls', 'uid': 'some_unique_id'}


def test_invalid_container():
    c = ContainerReference(TESTING_CONFIG['host'], TESTING_CONFIG['port'])
    inv_cont = dict(name='romans', uid=str(uuid.uuid4()),
                    time=str(ttime.time()), owner='obelix', project='invadegauls',
                    container=None,
                    beamline_id='trial_b')
    pytest.raises(HTTPError, c.create, **inv_cont)


def test_find_container():
    c = ContainerReference(TESTING_CONFIG['host'], TESTING_CONFIG['port'])
    f_cont = dict(name='comp_sam', uid=str(uuid.uuid4()),
                    time=ttime.time(), owner='hidefix', project='ceasar',
                    container='gauls',
                    beamline_id='fiction')
    c_uid = c.create(**f_cont)
    c_ret = next(c.find(uid=c_uid))
    assert c_ret['uid'] == c_uid


def test_find_container_as_doc():
    c = ContainerReference(TESTING_CONFIG['host'], TESTING_CONFIG['port'])
    f_cont = dict(name='comp_sam', uid=str(uuid.uuid4()),
                    time=ttime.time(), owner='hidefix', project='ceasar',
                    container='gauls',
                    beamline_id='fiction')
    c_uid = c.create(**f_cont)
    c_ret_doc = next(c.find(uid=c_uid, as_document=True))
    assert Document == type(c_ret_doc)
    assert c_ret_doc['uid'] == c_uid


def test_update_container():
    orig_cont = dict(name='obelix', uid=str(uuid.uuid4()),
                       time=ttime.time(), owner='chief', project='egypt',
                       beamline_id='fiction', state='active', container='SPQR')
    c = ContainerReference(host=TESTING_CONFIG['host'],
                              port=TESTING_CONFIG['port'])
    m_uid = c.create(**orig_cont)
    c.update(query={'uid': orig_cont['uid']},
             update={'state': 'inactive','time': time.time()})
    updated_cont = next(c.find(uid=m_uid))
    assert updated_cont['state'] == 'inactive'


def test_update_container_illegal():
    orig_cont = dict(name='obelix', uid=str(uuid.uuid4()),
                       time=ttime.time(), owner='chief', project='egypt',
                       beamline_id='fiction', state='active', container='SPQR')
    c = ContainerReference(host=TESTING_CONFIG['host'],
                              port=TESTING_CONFIG['port'])
    m_uid = c.create(**orig_cont)
    test_time = time.time()
    pytest.raises(HTTPError, c.update, query={'uid': orig_cont['uid']},
                  update={'uid': str(uuid.uuid4())})


def setup():
    amostra_setup()
    global sample_uids, document_insertion_time
    document_insertion_time = ttime.time()
    # TODO: Get sample data!!!!!! and populate



from tornado import gen
import tornado.web
import pymongo
import jsonschema
import ujson
from amostra.server import utils
from jsonschema.exceptions import ValidationError, SchemaError
from pymongo.errors import PyMongoError
from pymongo import DESCENDING
from .utils import compose_err_msg


def db_connect(database, mongo_host, mongo_port, mongo_user=None,
               mongo_pwd=None, auth=False):
    """Helper function to deal with stateful connections to MongoDB
    Connection established lazily. Connects to the database on request.
    Same connection pool is used for all clients per recommended by
    tornado developer manual.

    Parameters
    ----------
    database: str
        The name of database pymongo creates and/or connects
    host: str
        Name/address of the server that mongo daemon lives
    port: int
        Port num of the server
    Returns pymongo.database.Database
    -------
        Async server object which comes in handy as server has to juggle
    multiple clients and makes no difference for a single client compared
    to pymongo
    """
    if auth:
        uri = 'mongodb://{0}:{1}@{2}:{3}/'.format(mongo_user,
                                                  mongo_pwd,
                                                  mongo_host,
                                                  mongo_port)
        client = pymongo.MongoClient(uri)
    else:
        try:
            client = pymongo.MongoClient(host=mongo_host, port=mongo_port)
        except pymongo.errors.ConnectionFailure:
            raise utils.AmostraException("Unable to connect to MongoDB server...")
    database = client[database]
    try:
        database.sample.create_index([('uid', DESCENDING)],
                                     unique=True, background=True)
        database.sample.create_index([('time', DESCENDING),
                                     ('name', DESCENDING)],
                                     unique=False, background=True)
        database.sample.create_index([('container', DESCENDING)],
                                     unique=False, sparse=True)
        database.sample.create_index([('uid', DESCENDING)],
                                     unique=True, background=True)
        database.sample.create_index([('time', DESCENDING)],
                                     unique=False, background=True)
        database.request.create_index([('uid', DESCENDING)],
                                      unique=True, background=True)
        database.request.create_index([('time', DESCENDING)],
                                      unique=False, background=True)
        database.request.create_index([('sample', DESCENDING)],
                                      unique=False, background=True, sparse=True)
        database.container.create_index([('uid', DESCENDING)],
                                        unique=True, background=True)
        database.container.create_index([('time', DESCENDING)],
                                        unique=False, background=True)
        database.container.create_index([('container', DESCENDING)],
                                        unique=False, background=True,
                                        sparse=True)
    except PyMongoError:
        raise compose_err_msg(500, 'Not connected to Mongo daemon')
    return database


class DefaultHandler(tornado.web.RequestHandler):
    """DefaultHandler which takes care of CORS. Javascript needs this.
    In general, methods on RequestHandler and elsewhere in Tornado
    are not thread-safe. In particular, methods such as write(),
    finish(), and flush() must only be called from the main thread.
    If you use multiple threads it is important to use IOLoop.add_callback
    to transfer control back to the main thread before finishing the request.
    """
    @gen.coroutine
    def set_default_headers(self):
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Methods', 'POST, GET, OPTIONS')
        self.set_header('Access-Control-Max-Age', 1000)
        self.set_header('Access-Control-Allow-Headers', '*')
        self.set_header('Content-type', 'application/json')

    def data_received(self, chunk):
        """Abstract method, overwrite potential default"""
        pass


class SampleReferenceHandler(DefaultHandler):
    """Handler for SampleReference insert, update, and querying.
    A RESTful handler, nothing fancy or stateful about this

    Methods
    --------
    get()
        Query 'sample_referenece' documents given certain parameters.
    Pass 'num' in order to obtain the last 'num' sample references.

    post()
        Insert 'sample' documents

    put()
        Update or Insert if 'sample' document does not exist.
    Update supports both field update and document replacement. If you
    would like to replace a document, simply provide a full doc in update
    field. Otherwise, provide a dict that holds the new value and field name.
    Returns the total number of documents that are updated.
    """
    @gen.coroutine
    def get(self):
        database = self.settings['db']
        query = utils.unpack_params(self)
        num = query.pop("num", None)
        if num:
            try:
                docs = database.sample.find().sort('time',
                                                   direction=DESCENDING).\
                                                             limit(num)
            except pymongo.errors.PyMongoError:
                raise compose_err_msg(500, 'Query on sample has failed',
                                      query)
        else:
            try:
                docs = database.sample.find(query).sort('time',
                                                        direction=DESCENDING)
            except pymongo.errors.PyMongoError:
                raise compose_err_msg(500, 'Query Failed: ', query)
        if docs:
            utils.return2client(self, docs)
        else:
            raise compose_err_msg(500, 'No results found!')

    @gen.coroutine
    def post(self):
        database = self.settings['db']
        data = ujson.loads(self.request.body.decode("utf-8"))
        uids = []
        if isinstance(data, list):
            for d in data:
                d = utils.default_timeuid(d)
                try:
                    jsonschema.validate(d,
                                        utils.schemas['sample'])
                except (ValidationError, SchemaError):
                    raise compose_err_msg(400,
                                          "Invalid schema on document(s)",
                                          d)
                uids.append(d['uid'])
                res = database.sample.insert(d)
        elif isinstance(data, dict):
            data = utils.default_timeuid(data)
            try:
                jsonschema.validate(data,
                                    utils.schemas['sample'])
            except (ValidationError, SchemaError):
                raise compose_err_msg(400,
                                      "Invalid schema on document(s)",
                                      data)
            uids.append(data['uid'])
            res = database.sample.insert(data)
            if not res:
                raise compose_err_msg(500,
                                      'SampleHandler expects list or dict')
        self.finish(ujson.dumps(uids))

    @gen.coroutine
    def put(self):
        database = self.settings['db']
        incoming = ujson.loads(self.request.body)
        try:
            query = incoming.pop('query')
            update = incoming.pop('update')
        except KeyError:
            raise compose_err_msg(500,
                                  'filter and update are both required fields')
        if any(x in update.keys() for x in ['uid']):
            raise compose_err_msg(500,
                                  'Uid cannot be updated')
        res = database.sample.update_many(filter=query,
                                          update={'$set': update},
                                          upsert=False)
        self.finish(ujson.dumps(res.raw_result))


class RequestReferenceHandler(DefaultHandler):
    @gen.coroutine
    def get(self):
        database = self.settings['db']
        query = utils.unpack_params(self)
        num = query.pop("num", None)
        if num:
            try:
                docs = database.reference.find().\
                                          sort('time',
                                               direction=pymongo.DESCENDING).\
                                                         limit(num)
            except pymongo.errors.PyMongoError:
                raise utils._compose_err_msg(500, '', query)
        else:
            try:
                docs = database.request.find(query).sort('time',
                                                         direction=DESCENDING)
            except pymongo.errors.PyMongoError:
                raise utils._compose_err_msg(500, 'Query Failed: ', query)
        if docs:
            utils.return2client(self, docs)
        else:
            raise utils._compose_err_msg(500, 'No results found!')

    @gen.coroutine
    def post(self):
        database = self.settings['db']
        data = ujson.loads(self.request.body.decode("utf-8"))
        uids = []
        if isinstance(data, list):
            for d in data:
                d = utils.default_timeuid(d)
                try:
                    jsonschema.validate(d,
                                        utils.schemas['request'])
                except (ValidationError, SchemaError):
                    raise compose_err_msg(400,
                                          "Invalid schema on document(s)",
                                          d)
                try:
                    database.request.insert(d)
                    uids.append(d['uid'])
                except pymongo.errors.PyMongoError:
                    raise compose_err_msg(500,
                                         'Validated data can not be inserted',
                                                 data)
        elif isinstance(data, dict):
            data = utils.default_timeuid(data)
            try:
                jsonschema.validate(data,
                                    utils.schemas['request'])
            except (ValidationError, SchemaError):
                raise compose_err_msg(400,
                                      "Invalid schema on document(s)",
                                      data)
            try:
                database.request.insert(data)
                uids.append(data['uid'])
            except pymongo.errors.PyMongoError:
                raise compose_err_msg(500,
                                      'Validated data can not be inserted',
                                      data)
        else:
            raise compose_err_msg(500,
                                  status='SampleHandler expects list or dict')
        self.finish(ujson.dumps(uids))

    @gen.coroutine
    def put(self):
        database = self.settings['db']
        incoming = ujson.loads(self.request.body)
        try:
            query = incoming.pop('query')
            update = incoming.pop('update')
        except KeyError:
            raise compose_err_msg(500,
                                  'filter and update are both required fields')
        if any(x in update.keys() for x in ['uid']):
            raise compose_err_msg(500,
                                  status='Uid cannot be updated')
        res = database.request.update_many(filter=query,
                                           update={'$set': update},
                                           upsert=False)
        self.finish(ujson.dumps(res.raw_result))


class ContainerReferenceHandler(DefaultHandler):
    """Handler for SampleReference insert, update, and querying.
    A RESTful handler, nothing fancy or stateful about this

    Methods
    --------
    get()
        Query 'container' documents given certain parameters.
    Pass 'num' in order to obtain the last 'num' sample references.

    post()
        Insert 'sample' documents

    put()
        Update or Insert if 'sample' document does not exist.
    Update supports both field update and document replacement. If you
    would like to replace a document, simply provide a full doc in update
    field. Otherwise, provide a dict that holds the new value and field name.
    Returns the total number of documents that are updated.
    """
    @gen.coroutine
    def get(self):
        database = self.settings['db']
        query = utils.unpack_params(self)
        num = query.pop("num", None)
        if num:
            try:
                docs = database.sample.find().sort('time',
                                                   direction=DESCENDING).\
                                                             limit(num)
            except pymongo.errors.PyMongoError:
                raise compose_err_msg(500, '', query)
        else:
            try:
                docs = database.container.find(query).sort('time',
                                                      direction=DESCENDING)
            except pymongo.errors.PyMongoError:
                raise compose_err_msg(500, 'Query Failed: ', query)
        if docs:
            utils.return2client(self, docs)
        else:
            raise compose_err_msg(500, 'No results found!')

    @gen.coroutine
    def post(self):
        database = self.settings['db']
        data = ujson.loads(self.request.body.decode("utf-8"))
        uids = []
        if isinstance(data, list):
            for d in data:
                d = utils.default_timeuid(d)
                try:
                    jsonschema.validate(d,
                                        utils.schemas['container'])
                except (ValidationError, SchemaError):
                    raise compose_err_msg(400,
                                          "Invalid schema on document(s)", d)
                uids.append(d['uid'])
                res = database.container.insert(d)
        elif isinstance(data, dict):
            data = utils.default_timeuid(data)
            try:
                jsonschema.validate(data,
                                    utils.schemas['container'])
            except (ValidationError, SchemaError):
                raise compose_err_msg(400,
                                      "Invalid schema on document(s)", data)
            uids.append(data['uid'])
            res = database.container.insert(data)
            if not res:
                raise compose_err_msg(500,
                                      'SampleHandler expects list or dict')
        self.finish(ujson.dumps(uids))

    @gen.coroutine
    def put(self):
        database = self.settings['db']
        incoming = ujson.loads(self.request.body)
        try:
            query = incoming.pop('query')
            update = incoming.pop('update')
        except KeyError:
            raise compose_err_msg(500,
                                  'filter and update are both required fields')
        if any(x in update.keys() for x in ['uid']):
            raise compose_err_msg(500,
                                  'Uid cannot be updated')
        res = database.container.update_many(filter=query,
                                             update={'$set': update},
                                             upsert=False)
        self.finish(ujson.dumps(res.raw_result))


class SchemaHandler(DefaultHandler):
    """Provides the json used for schema validation provided collection name"""
    @gen.coroutine
    def get(self):
        col = utils.unpack_params(self)
        self.write(utils.schemas[col])
        self.finish()

    @gen.coroutine
    def put(self):
        raise utils._compose_err_msg(405,
                                     status='Not allowed on server')

    @gen.coroutine
    def post(self):
        raise utils._compose_err_msg(405,
                                     status='Not allowed on server')



import time as ttime
import pytest
import time
from amostra.testing import amostra_setup, amostra_teardown
from amostra.client.api import RequestReference
from requests.exceptions import HTTPError, ConnectionError
from amostra.testing import TESTING_CONFIG
from uuid import uuid4
from requests.exceptions import ConnectionError

def teardown():
    amostra_teardown()


def test_request_constructor_config():
    r1 = RequestReference(host=TESTING_CONFIG['host'],
                          port=TESTING_CONFIG['port'])
    assert r1.host == TESTING_CONFIG['host']
    assert r1.port == TESTING_CONFIG['port']


def test_request_create():
    req1= RequestReference(host=TESTING_CONFIG['host'],
                           port=TESTING_CONFIG['port'])

    req2 = RequestReference(host=TESTING_CONFIG['host'],
                           port=TESTING_CONFIG['port'])
    req2.create(sample='roman_sample', time=time.time(),
               uid=None, state='active', seq_num=0, foo='bar',
               hero='asterix', antihero='romans')


def test_duplicate_request():
    m_uid = str(uuid4())
    req1= RequestReference(host=TESTING_CONFIG['host'],
                           port=TESTING_CONFIG['port'])
    req1.create(sample='hidefix', time=time.time(),
               uid=m_uid, state='active', seq_num=0, foo='bar',
               hero='asterix', antihero='romans')
    req2 = RequestReference(host=TESTING_CONFIG['host'],
                           port=TESTING_CONFIG['port'])
    pytest.raises(HTTPError, req2.create, sample='hidefix', time=time.time(),
               uid=m_uid, state='active', seq_num=0, foo='bar',
               hero='asterix', antihero='romans')

def test_request_find():
    r = RequestReference(host=TESTING_CONFIG['host'],
                        port=TESTING_CONFIG['port'])
    req_dict = dict(sample='hidefix', time=time.time(),
                    uid=str(uuid4()), state='active', seq_num=0, foo='bar',
                    hero='obelix', antihero='romans')
    inserted = r.create(**req_dict)
    retrieved = next(r.find(foo='bar'))
    assert retrieved['uid'] == inserted


def test_update_request():
    r = RequestReference(host=TESTING_CONFIG['host'],
                        port=TESTING_CONFIG['port'])
    m_uid = str(uuid4())
    req_dict = dict(sample='hidefix', time=time.time(),
                    uid=m_uid, state='active', seq_num=0, foo='bar',
                    hero='obelix', antihero='romans')
    r.create(**req_dict)
    r.update(query={'uid': m_uid},
                       update={'state': 'inactive'})
    updated_req = next(r.find(uid=m_uid))
    assert updated_req['state'] == 'inactive'


def setup():
    amostra_setup()
    global sample_uids, document_insertion_time
    document_insertion_time = ttime.time()
    # TODO: Get sample data!!!!!! and populate



import tornado.web
from pkg_resources import resource_filename as rs_fn
import ujson
import pymongo
import uuid
import time as ttime

class AmostraException(Exception):
    pass


SCHEMA_PATH = 'schemas'
SCHEMA_NAMES = {'sample': 'sample.json',
                'request': 'request.json',
                'container': 'container.json'}
fn = '{}/{{}}'.format(SCHEMA_PATH)
schemas = {}
for name, filename in SCHEMA_NAMES.items():
    try:
        with open(rs_fn('amostra',
                        resource_name=fn.format(filename))) as fin:
            schemas[name] = ujson.load(fin)
    except FileNotFoundError:
        raise AmostraException('Schema file not found or does not exist')


def compose_err_msg(code, status, m_str=''):
    fmsg = str(status) + str(m_str)
    return tornado.web.HTTPError(status_code=code, reason=fmsg)


def unpack_params(handler):
    """Unpacks the queries from the body of the header
    Parameters
    ----------
    handler: tornado.web.RequestHandler
        Handler for incoming request to collection
    Returns dict
    -------
        Unpacked query in dict format.
    """
    if isinstance(handler, tornado.web.RequestHandler):
        return ujson.loads(list(handler.request.arguments.keys())[0])
    else:
        raise TypeError("Wrong type")


def return2client(handler, payload):
    """Home brew solution to dump the result back to client's open socket.
    No need to worry about package size or socket behavior as
    tornado handles this for us
    Parameters
    -----------
    handler: tornado.web.RequestHandler
        Request handler for the collection of operation(post/get)
    payload: dict, list
        Information to be sent to the client
    """
    # TODO: Solve precision issue with json precision
    if isinstance(payload, pymongo.cursor.Cursor):
            l = []
            for p in payload:
                del(p['_id'])
                l.append(p)
            handler.write(ujson.dumps(l))
    elif isinstance(payload, dict):
        del(payload['_id'])
        handler.write(ujson.dumps(list(payload)))
    else:
        handler.write('[')
        d = next(payload)
        while True:
            try:
                del(d['_id'])
                handler.write(ujson.dumps(d))
                d = next(payload)
                handler.write(',')
            except StopIteration:
                break
        handler.write(']')
    handler.finish()


def default_timeuid(document):
    """
    Given a document, default time and uid fields if not provided

    Parameters
    ----------
    document: dict
        Document to be inserted

    Returns
    -------
    dict
        Document with defaults
    """
    if 'uid' not in document or document['uid'] is None:
        document['uid'] = str(uuid.uuid4())
    if 'time' not in document or document['time'] is None:
        document['time'] = ttime.time()
    return document



from amostra.testing import _baseSM, TESTING_CONFIG
from amostra.client.commands import SampleReference


class TestBasicSampleRef(_baseSM):
    db_class = SampleReference
    args = ()
    kwargs = {}
    db_class.host = TESTING_CONFIG['host']
    db_class.port = TESTING_CONFIG['port']

    def test_disk_round_trip(self):
        db = self.db_class(*self.args, **self.kwargs)

        def _helper(doc_list):
            db2 = SampleReference(doc_list)
            for doc in db.find():
                doc_id = doc['uid']
                rl_doc = db2.find(uid=doc_id)
                assert doc == rl_doc



from doct import Document
import time as ttime
import pytest
from ..testing import amostra_setup, amostra_teardown
from ..client.api import AmostraClient
from requests.exceptions import HTTPError, RequestException
from ..testing import TESTING_CONFIG

import uuid


conn = AmostraClient(host=TESTING_CONFIG['host'],
                         port=TESTING_CONFIG['port'])


def teardown():
    amostra_teardown()

def test_client_constructor():
    pytest.raises(TypeError, AmostraClient)


def test_connection_switch(conn=conn):
    conn.host = 'caesar'
    pytest.raises(RequestException, conn.create_sample, 'asterix')
    conn.host = TESTING_CONFIG['host']


def test_sample_create(conn=conn):
   r1 = conn.create_sample(name='test1')
   r2 = conn.create_sample(name='test2', uid=str(uuid.uuid4()))
   m_kwargs = dict(owner='test', level='inferno', type='automated',
                   material='CoFeB')
   r3 = conn.create_sample(name='test3', uid=str(uuid.uuid4()), **m_kwargs)


def test_duplicate_sample(conn=conn):
    _com_uid = str(uuid.uuid4())
    conn.create_sample(name='test_sample', uid=_com_uid, custom=False)
    pytest.raises(HTTPError, conn.create_sample, name='test_duplicate',
                  uid=_com_uid)


def test_invalid_sample(conn=conn):
    pytest.raises(TypeError, conn.create_sample)


def test_find_sample(conn=conn):
    m_sample = dict(name='comp_samp', uid=str(uuid.uuid4()),
                    time=ttime.time(), owner='test', project='ci-tests',
                    beamline_id='test-ci')
    s = conn.create_sample(**m_sample)
    s_ret = conn.find_sample(uid=m_sample['uid'])
    assert s_ret[0]['uid'] == m_sample['uid']


def test_find_sample_as_doc(conn=conn):
    m_sample = dict(name='comp_sam', uid=str(uuid.uuid4()),
                    time=ttime.time(), owner='arkilic', project='trial',
                    beamline_id='trial_b', container='legion1')
    conn.create_sample(**m_sample)
    s_ret = conn.find_sample(uid=m_sample['uid'], as_document=True)
    assert s_ret[0] == Document('Sample', m_sample)

def test_update_sample(conn=conn):
    test_sample = dict(name='up_sam', uid=str(uuid.uuid4()),
                       time=ttime.time(), owner='arkilic', project='trial',
                       beamline_id='trial_b', state='active', container='legion2')
    conn.create_sample(**test_sample)
    conn.update_sample(query={'uid': test_sample['uid']},
                       update={'state': 'inactive', 'time': ttime.time()})
    updated_samp = conn.find_sample(name='up_sam')[0]
    assert updated_samp['state'] == 'inactive'


def test_update_sample_illegal(conn=conn):
    test_sample = dict(name='up_sam', uid=str(uuid.uuid4()),
                    time=ttime.time(), owner='arkilic', project='trial',
                    beamline_id='trial_b', updated=False)
    pytest.raises(HTTPError,
                  conn.update_sample, query={'name': test_sample['name']},
                                      update={'uid': 'illegal'})

def test_container_create(conn=conn):
    ast_cont = {'name': 'obelix', "dog": 'hidefix', 'time': ttime.time(),
                'container': 'gauls', 'uid': str(uuid.uuid4())}
    cont1 = conn.create_container(**ast_cont)
    assert cont1 == ast_cont['uid']


def test_find_container(conn=conn):
    f_cont = dict(name='comp_sam', uid=str(uuid.uuid4()),
                    time=ttime.time(), owner='hidefix', project='ceasar',
                    container='gauls',
                    beamline_id='fiction')
    c_uid = conn.create_container(**f_cont)
    c_ret_doc = conn.find_container(uid=c_uid, as_document=True)[0]
    assert Document == type(c_ret_doc)
    assert c_ret_doc['uid'] == c_uid


def test_find_container_as_doc(conn=conn):
    f_cont = dict(name='comp_sam', uid=str(uuid.uuid4()),
                    time=ttime.time(), owner='hidefix', project='ceasar',
                    container='gauls',
                    beamline_id='fiction')
    c_uid = conn.create_container(**f_cont)
    c_ret_doc = conn.find_container(uid=c_uid, as_document=True)[0]
    assert Document == type(c_ret_doc)
    assert c_ret_doc['uid'] == c_uid

def test_update_container(conn=conn):
    orig_cont = dict(name='obelix', uid=str(uuid.uuid4()),
                       time=ttime.time(), owner='chief', project='egypt',
                       beamline_id='fiction', state='active', container='SPQR')
    m_uid = conn.create_container(**orig_cont)
    conn.update_container(query={'uid': orig_cont['uid']},
                          update={'state': 'inactive','time': ttime.time()})
    updated_cont = conn.find_container(uid=m_uid)[0]
    assert updated_cont['state'] == 'inactive'


def test_update_container_illegal(conn=conn):
    orig_cont = dict(name='obelix', uid=str(uuid.uuid4()),
                       time=ttime.time(), owner='chief', project='egypt',
                       beamline_id='fiction', state='active', container='SPQR')
    m_uid = conn.create_container(**orig_cont)
    test_time = ttime.time()
    pytest.raises(HTTPError, conn.update_container, query={'uid': orig_cont['uid']},
                  update={'uid': str(uuid.uuid4())})


def test_request_create(conn=conn):
    conn.create_request(sample='roman_sample', time=ttime.time(),
               uid=None, state='active', seq_num=0, foo='bar',
               hero='asterix', antihero='romans')


def test_duplicate_request(conn=conn):
    m_uid = str(uuid.uuid4())
    conn.create_request(sample='hidefix', time=ttime.time(),
               uid=m_uid, state='active', seq_num=0, foo='bar',
               hero='asterix', antihero='romans')
    pytest.raises(HTTPError, conn.create_request, sample='hidefix', time=ttime.time(),
               uid=m_uid, state='active', seq_num=0, foo='bar',
               hero='asterix', antihero='romans')

def test_request_find(conn=conn):
    req_dict = dict(sample='hidefix', time=ttime.time(),
                    uid=str(uuid.uuid4()), state='active', seq_num=0, foo='bar',
                    hero='obelix', antihero='romans')
    inserted = conn.create_request(**req_dict)
    retrieved = conn.find_request(uid=inserted)[0]
    assert retrieved['uid'] == inserted


def test_update_request(conn=conn):
    m_uid = str(uuid.uuid4())
    req_dict = dict(uid=m_uid, sample='hidefix', time=ttime.time(),
                    seq_num=0, foo='bar',
                    hero='obelix', antihero='romans')
    conn.create_request(**req_dict)
    conn.update_request(query={'uid': m_uid},
                        update={'state': 'inactive'})
    updated_req = conn.find_request(uid=m_uid)[0]
    assert updated_req['state'] == 'inactive'



import argparse
import tornado.web
import sys
import tornado.ioloop
import tornado.options
from  amostra.server.engine import (SampleReferenceHandler,
                                    RequestReferenceHandler,
                                    SchemaHandler,
                                    ContainerReferenceHandler,
                                    db_connect)
from .server.conf import load_configuration


def start_server(config=None):
    """
    Amostra service startup script.
    Returns tornado event loop provided configuration.

    Parameters
    ----------
    config: dict
        Command line arguments always have priority over local config or
        yaml files.Using these parameters,  a tornado event loop is created.
        Keep in mind that this server is started in lazy fashion.
        It does not verify
        the existence of a mongo instance running on the specified location.
    """
    if not config:
        config = {k: v for k, v in load_configuration('amostra', 'AMST',
                                                      ['mongo_host',
                                                       'mongo_port', 'timezone',
                                                       'database',
                                                       'service_port'],
                                                      allow_missing=True).items() if v is not None}
    parser = argparse.ArgumentParser()
    parser.add_argument('--database', dest='database', type=str,
                        help='name of database to use')
    parser.add_argument('--mongo-host',
                         dest='mongo_host', type=str,
                        help='host to use')
    parser.add_argument('--timezone', dest='timezone', type=str,
                        help='Local timezone')
    parser.add_argument('--mongo-port', dest='mongo_port', type=int,
                        help='port to use to talk to mongo')
    parser.add_argument('--service-port', dest='service_port', type=int,
                        help='port listen to for clients')
    parser.add_argument('--no-auth', dest='auth', action='store_false')
    parser.add_argument('--auth', dest='auth', action='store_true')
    parser.set_defaults(auth=False)
    parser.add_argument('--mongo-user', dest='mongo_user', type=str,
                            help='Mongo username')
    parser.add_argument('--mongo-pwd', dest='mongo_pwd', type=str,
                            help='Mongo password')
    args = parser.parse_args()
    print(args)
    if args.database is not None:
        config['database'] = args.database
    if args.mongo_host is not None:
        config['mongo_host'] = args.mongo_host
    if args.timezone is not None:
        config['timezone'] = args.timezone
    if args.mongo_port is not None:
        config['mongo_port'] = args.mongo_port
    service_port = args.service_port
    if service_port is None:
        service_port = 7770
    if args.auth:
        if args.mongo_user and args.mongo_pwd:
            config['mongo_user'] = args.mongo_user
            config['mongo_pwd'] = args.mongo_pwd
        else:
            raise KeyError('--mongo-user and --mongo-pwd required with auth')
    else:
        config['mongo_user'] = None
        config['mongo_pwd'] = None
    db = db_connect(database=config['database'],
                    mongo_host=config['mongo_host'],
                    mongo_port=config['mongo_port'],
                    mongo_user=config['mongo_user'],
                    mongo_pwd=config['mongo_pwd'],
                    auth=args.auth)
    application = tornado.web.Application([
        (r'/sample', SampleReferenceHandler),
        (r'/request', RequestReferenceHandler),
        (r'/container', ContainerReferenceHandler),
        (r'/schema', SchemaHandler)
         ], db=db)
    print('Starting Amostra service with configuration ', config)
    application.listen(service_port)
    tornado.ioloop.IOLoop.current().start()



from doct import Document
import ujson
import requests
from ..client import conf
from .amutils import doc_or_uid_to_uid, _get, _post, _put


class AmostraException(Exception):
    pass


class AmostraClient:
    def __init__(self, host, port):
        """ Simplified client for Sample, Request, and Container

        Parameters
        ----------
        host: str, optional
            Machine name/address for Amostra server
        port: int, optional
            Port Amostra server is initiated on
        """
        self.host = host
        self.port = port

    @property
    def _sample_client(self):
        """ Connection pool for Sample related tasks."""
        return SampleReference(host=self.host,
                               port=self.port)

    @property
    def _container_client(self):
        """Connection pool for Container related tasks."""
        return ContainerReference(host=self.host,
                                  port=self.port)

    @property
    def _request_client(self):
        """Connection pool for Request related tasks."""
        return RequestReference(host=self.host,
                                port=self.port)

    def create_sample(self, name, time=None, uid=None, container=None,
                       **kwargs):
        """ Insert a sample to the database

        Parameters
        ----------
        name : str
           The name of the sample  This must be unique in the database
        time: float
           The time sample is created. Client provided timestamp
        uid: str
            Unique identifier for a sample document
        container: str, list
            A single or a list of container documents or uids

        Returns
        -------
        str: The inserted document.

        """
        return self._sample_client.create(name=name, time=time, uid=uid,
                                          container=container, **kwargs)

    def create_request(self, sample=None, time=None,
                       uid=None, state='active', seq_num=0, **kwargs):
        """ Create a request entry in the dataase

        Parameters
        ----------
        sample: dict, doct.Document, optional
            The sample this reference refers to
        time: float
            Time request got created
        uid: str
            Unique identifier for
        state: str
            Enum 'active' or 'inactive' that displays the state of a request
        seq_num: int
            Sequence number for creation of the request. Not indexed but can be updated

        Returns
        -------
        str: The inserted Request document uid

        """
        return self._request_client.create(sample=sample, time=time, uid=uid,
                                           state=state, seq_num=seq_num,
                                           **kwargs)

    def create_container(self, uid=None, time=None, **kwargs):
        """ Insert a container document.

        Parameters
        ----------
        uid: str
        Unique identifier for a Container document
        time: float
        Time document created. Client side timestamp

        Returns
        -------
        str: Inserted Container document uid

        """
        return self._container_client.create(uid=uid, time=time, **kwargs)

    def update_sample(self, update, query):
        """Update a sample given a query and name value pair to be updated.
        No upsert(s).

        Parameters
        -----------
        query: dict
            Allows finding Request documents to be updated.
        update: dict
            Name/value pair that is to be replaced within an existing Request doc.
        Returns
        ----------
        bool
            Returns True if update successful
        """
        return self._sample_client.update(update=update, query=query)

    def update_request(self, update, query):
        """Update a request given a query and name value pair to be updated.
        No upsert(s).

        Parameters
        -----------
        query: dict
        Allows finding Request documents to be updated.
        update: dict
        Name/value pair that is to be replaced within an existing Request doc.
        Returns
        ----------
        bool
        Returns True if update successful

        """
        if self._request_client.update(update=update, query=query):
            return True
        return False

    def update_container(self, update, query):
        """Update a container given a query and name value pair to be updated.
        No upsert(s).

        Parameters
        -----------
        query: dict
            Allows finding Request documents to be updated.
        update: dict
            Name/value pair that is to be replaced within an existing Request doc.
        Returns
        ----------
        bool
            Returns True if update successful

        """
        if self._container_client.update(update=update, query=query):
            return True
        return False

    def find_sample(self, as_document=True, **kwargs):
        """Given a set of mongo search parameters, return a requests iterator

        Parameters
        -----------
        as_document: bool
            Format return type to doct.Document if set

        Yields
        ----------
       list
            Result of the query, list of samples doct.Document


        """
        return list(self._sample_client.find(as_document=as_document,
                                                 **kwargs))

    def find_request(self, as_document=True, **kwargs):
        """Given a set of mongo search parameters, return a list of requests

        Parameters
        -----------
        as_document: bool
            Format return type to doct.Document if set

        Returns
        ----------
        list
            Result of the query, list of doct.Document

        """
        return list(self._request_client.find(as_document=as_document,
                                              **kwargs))

    def find_container(self, as_document=True, **kwargs):
        """Given a set of mongo search parameters, return a requests iterator

        Parameters
        -----------
        as_document: bool
            Format return type to doct.Document if set

        Yields
        ----------
       list
            Result of the query, list of containers doct.Document

        """
        return list(self._container_client.find(as_document=as_document,
                                                **kwargs))

class SampleReference(object):
    """Reference implementation of generic sample manager"""
    def __init__(self, host=conf.conn_config['host'],
                 port=conf.conn_config['port']):
        """Constructor.

        Parameters
        ----------
        host: str, optional
            Machine name/address for Amostra server
        port: int, optional
            Port Amostra server is initiated on

        """
        self.host = host
        self.port = port

    @property
    def _server_path(self):
        """URL to the Amostra server"""
        return 'http://{}:{}/' .format(self.host, self.port)

    @property
    def _samp_url(self):
        """URL to the sample reference handler in the server side"""
        return self._server_path + 'sample'

    @property
    def _schema_url(self):
        """URL to the schema reference handler in the server side"""
        return self._server_path + 'schema'

    def create(self, name, time=None, uid=None, container=None,
               **kwargs):
        """Insert a sample to the database

        Parameters
        ----------
        name : str
            The name of the sample.  This must be unique in the database
        time: float
            The time sample is created. Client provided timestamp
        uid: str
            Unique identifier for a sample document
        container: str, list
            A single or a list of container documents or uids

        Returns
        -------
        ins_doc : str
            The inserted document uid

        """
        doc = dict(uid=uid,
                   name=name, time=time,
                   container=doc_or_uid_to_uid(container) if container else 'NULL',
                   **kwargs)
        ins_doc = _post(self._samp_url, doc)
        return ins_doc[0]

    def create_sample_list(self, sample_list):
        """Insert a sample to the database

        Parameters
        ----------
        sample_list : list
            List of dictionaries

        Returns
        -------
        uid_list : list
            uid of the inserted documents

        """
        uid_list = []
        for s in sample_list:
            uid = _post(self._samp_url, s)
            uid_list.extend(uid)
        return uid_list

    def update(self, query, update):
        """Update a request given a query and name value pair to be updated. No upsert support.
        For more info on upsert, check Mongo documentations

        Parameters
        -----------
        query: dict
            Allows finding Sample documents to be updated
        update: dict
            Name/value pair that is to be replaced within an existing Request doc
        Returns
        ----------
        bool
            Returns True if update successful

        """
        _put(self._samp_url, query, update)
        return True

    def find(self, as_document=False, **kwargs):
        """
        Parameters
        ----------
        as_document: bool
            Formats output to doct.Document if True

        Yields
        ------
        c : dict, doct.Document
            Documents which have all keys with the given values

        """
        content = _get(self._samp_url, params=kwargs)
        if as_document:
            for c in content:
                yield Document('Sample', c)
        else:
            for c in content:
                yield c

    def get_schema(self):
        """Get information about schema from the server side

        Returns
        --------
        dict
            Returns the json schema dict used for validation

        """
        r = requests.get(self._schema_url,
                         params=ujson.dumps('sample'))
        r.raise_for_status()
        return ujson.loads(r.text)


class RequestReference(object):
    """Reference implementation of generic request"""
    def __init__(self, host=conf.conn_config['host'], port=conf.conn_config['port']):
        """Constructor

        Parameters
        ----------
        host: str, optional
            Machine name/address for amostra server
        port: int, optional
            Port amostra server is initiated on

        """
        self.host = host
        self.port = port

    @property
    def _server_path(self):
        """URL to the Amostra server"""
        return 'http://{}:{}/' .format(self.host, self.port)

    @property
    def _req_url(self):
        """URL to the request reference handler in the server side"""
        return self._server_path + 'request'

    @property
    def _schema_url(self):
        """URL to the schema reference handler in the server side"""
        return self._server_path + 'schema'

    def create(self, sample=None, time=None,
               uid=None, state='active', seq_num=0, **kwargs):
        """ Create a request entry in the dataase
        Parameters
        ----------
        sample: dict, doct.Document, optional
            The sample this reference refers to
        time: float
            Time request got created
        uid: str
            Unique identifier for
        state: str
            Enum 'active' or 'inactive' that displays the state of a request
        seq_num: int
            Sequence number for creation of the request. Not indexed but can be updated

        Returns
        -------
        ins_doc: str
            The inserted Request document uid

        """

        payload = dict(uid=uid,
                       sample=doc_or_uid_to_uid(sample),
                       time=time, state=state,
                       seq_num=seq_num, **kwargs)
        ins_doc = _post(self._req_url, payload)
        return ins_doc[0]

    def find(self, as_document=False, **kwargs):
        """Given a set of mongo search parameters, return a requests iterator

        Parameters
        -----------
        as_document: bool
            Format return type to doct.Document if set

        Yields
        ----------
        dict, doct.Document
            Result of the query

        Raises
        ---------
        StopIteration, requests.exceptions.HTTPError
            When nothing found or something is wrong on the server side. If server error occurs,
            a human friendly message is returned.

        """
        content = _get(self._req_url, kwargs)
        if as_document:
            for c in content:
                yield Document('Request', c)
        else:
            for c in content:
                yield c

    def update(self, query, update):
        """Update a request given a query and name value pair to be updated.
        No upsert(s).

        Parameters
        -----------
        query: dict
            Allows finding Request documents to be updated.
        update: dict
            Name/value pair that is to be replaced within an existing Request doc.
        Returns
        ----------
        bool
            Returns True if update successful
        """
        _put(self._req_url, query, update)
        return True

    def get_schema(self):
        """Get information about schema from the server side

        Returns
        --------
        dict
            Returns the json schema dict used for validation
        """
        r = requests.get(self._schema_url,
                         params=ujson.dumps('request'))
        r.raise_for_status()
        return ujson.loads(r.text)


class ContainerReference(object):
    """Reference implementation of generic container"""
    def __init__(self, host=conf.conn_config['host'], port=conf.conn_config['port']):
        """Handles connection configuration to the service backend.
        Either initiate with a request or use purely as a client for requests.
        """
        self.port = port
        self.host = host

    @property
    def _server_path(self):
        """URL to the Amostra server"""
        return 'http://{}:{}/' .format(self.host, self.port)

    @property
    def _cont_url(self):
        """URL to the container reference handler in the server side"""
        return self._server_path + 'container'

    @property
    def _schema_url(self):
        """URL to the schema reference handler in the server side"""
        return self._server_path + 'schema'

    def create(self, uid=None, time=None, **kwargs):
        """Insert a container document. Schema validation done
        on the server side. No native Python object (e.g. np.ndarray)
        due to performance constraints.

        Parameters
        ----------
        uid: str
            Unique identifier for a Container document
        time: float
            Time document created. Client side timestamp

        Returns
        -------
        ins_doc: str
            Inserted Container document uid
        """
        payload = dict(uid=uid,
                       time=time, **kwargs)
        ins_doc = _post(self._cont_url, payload)
        return ins_doc[0]

    def find(self, as_document=False, **kwargs):
        """Given a set of MongoDB search parameters, return a requests iterator"""
        content = _get(self._cont_url, kwargs)
        if as_document:
            for c in content:
                yield Document('container', c)
        else:
            for c in content:
                yield c

    def update(self, query, update):
        """Update a request given a query and name value pair to be updated.
        No upsert(s). If doc does not exist, simply do not update

        Parameters
        -----------
        query: dict
            Allows finding Container documents to be updated.
        update: dict
            Name/value pair that is to be replaced within an existing Request doc.
        Returns
        --------
        bool
            Returns True if update successful

        Returns
        ----------
        bool
            Returns True if update successful
        """
        _put(self._cont_url, query, update)
        return True

    def get_schema(self):
        """Get information about schema from the server side

        Returns
        --------
        dict
            Returns the json schema dict used for validation
        """
        r = requests.get(self._schema_url, params=ujson.dumps('container'))
        r.raise_for_status()
        return ujson.loads(r.text)



# Smoketest the api
from amostra.client.api import (SampleReference, RequestReference,
                                ContainerReference, AmostraClient)

def test_api_smoke():
    s_create = SampleReference.create
    s_find = SampleReference.find
    s_update = SampleReference.update

    r_create = RequestReference.create
    r_find = RequestReference.find
    r_update = RequestReference.update

    c_create = ContainerReference.create
    r_find = ContainerReference.find
    r_update = ContainerReference.update

    create_samp = AmostraClient.create_sample
    create_cont = AmostraClient.create_container
    create_cont = AmostraClient.create_request

    find_samp = AmostraClient.find_sample
    find_cont = AmostraClient.find_container
    find_req = AmostraClient.find_request

    update_samp = AmostraClient.update_sample
    update_cont = AmostraClient.update_container
    update_req = AmostraClient.update_request



from .commands import (SampleReference, RequestReference, ContainerReference)
from .local_commands import (LocalSampleReference, LocalRequestReference,
                             LocalContainerReference)
from .commands import AmostraClient



import requests
import ujson


def doc_or_uid_to_uid(doc_or_uid):
    """Given Document or uid return the uid

    Parameters
    ----------
    doc_or_uid : dict or str
        If str, then assume uid and pass through, if not, return
        the 'uid' field

    Returns
    -------
    uid : str
        A string version of the uid of the given document
    """
    try:
        return str(doc_or_uid['uid'])
    except TypeError:
        return doc_or_uid


def _get(url, params):
    """RESTful API get (querying)

    Parameters
    ----------
    url: str
        Address for the amostra server
    params: dict
        Query parameters to be sent to mongo instance

    Returns
    -------
    list
        Results of the query

    """
    r = requests.get(url, ujson.dumps(params))
    r.raise_for_status()
    return ujson.loads(r.text)


def _post(url, data):
    """RESTful API post (insert to database)

    Parameters
    ----------
    url: str
        Address for the amostra server
    data: dict
        Entries to be inserted to database

    """
    r = requests.post(url,
                      data=ujson.dumps(data))
    r.raise_for_status()
    return r.json()


def _put(url, query, update):
    """RESTful API put (update entries in database)

    Parameters
    ----------
    url: str
        Address for the amostra server
    query: dict
        Query string. Any pymongo supported mongo query
    update: dict
        Key/value pairs to be updated in the original document

    Returns
    -------
    bool
        True if update successful
    Raises
    --------
    requests.exceptions.HTTPError
        In case update fails, appropriate HTTPError and message string returned

    """
    update_cont = {'query': query, 'update': update}
    r = requests.put(url,
                     data=ujson.dumps(update_cont))
    r.raise_for_status()



from doct import Document
import time as ttime
import pytest

from ..testing import amostra_setup, amostra_teardown
from ..client.api import SampleReference, AmostraClient
from requests.exceptions import HTTPError, RequestException
from ..testing import TESTING_CONFIG

import uuid

sample_uids = []
document_insertion_times = []

def teardown():
    amostra_teardown()


def test_sample_constructor():
    s2 = SampleReference()


def test_connection_switch():
    s = SampleReference()
    s.host = 'caesar'
    pytest.raises(RequestException, s.create, 'asterix')
    s.host = TESTING_CONFIG['host']
    s.create(name='asterix')


def test_sample_create():
    samp = SampleReference()
    samp.host = TESTING_CONFIG['host']
    samp.port = TESTING_CONFIG['port']
    r1 = samp.create(name='test1')
    r2 = samp.create(name='test2', uid=str(uuid.uuid4()))
    m_kwargs = dict(owner='test', level='inferno', type='automated',
                    material='CoFeB')
    r3 = samp.create(name='test3', uid=str(uuid.uuid4()), **m_kwargs)


def test_duplicate_sample():
    s = SampleReference(host=TESTING_CONFIG['host'],
                        port=TESTING_CONFIG['port'])
    _comm_uid = str(uuid.uuid4())
    s.create(name='test_samp', uid=_comm_uid)
    pytest.raises(HTTPError, s.create, name='test_dup', uid=_comm_uid)


def test_invalid_sample():
    s = SampleReference(host=TESTING_CONFIG['host'],
                        port=TESTING_CONFIG['port'])
    pytest.raises(TypeError, s.create)


def test_find_sample():
    m_sample = dict(name='comp_sam', uid=str(uuid.uuid4()),
                    time=ttime.time(), owner='arkilic', project='trial',
                    container='test_group1',
                    beamline_id='trial_b')

    s1 = SampleReference(host=TESTING_CONFIG['host'],
                         port=TESTING_CONFIG['port'])
    s1.create(**m_sample)

    s_ret = next(s1.find(uid=m_sample['uid']))
    assert s_ret == m_sample


def test_find_sample_as_doc():
    m_sample = dict(name='comp_sam', uid=str(uuid.uuid4()),
                    time=ttime.time(), owner='arkilic', project='trial',
                    beamline_id='trial_b', container='legion1')
    s1 = SampleReference(host=TESTING_CONFIG['host'],
                         port=TESTING_CONFIG['port'])
    s1.create(**m_sample)
    s_ret = next(s1.find(uid=m_sample['uid'], as_document=True))
    assert s_ret == Document('Sample', m_sample)


def test_update_sample():
    test_sample = dict(name='up_sam', uid=str(uuid.uuid4()),
                       time=ttime.time(), owner='arkilic', project='trial',
                       beamline_id='trial_b', state='active', container='legion2')
    samp = SampleReference(host=TESTING_CONFIG['host'],
                           port=TESTING_CONFIG['port'])
    samp.create(**test_sample)
    samp.update(query={'uid': test_sample['uid']},
                update={'state': 'inactive', 'time': ttime.time()})
    updated_samp = next(samp.find(name='up_sam'))
    assert updated_samp['state'] == 'inactive'

def test_update_sample_illegal():
    test_sample = dict(name='up_sam', uid=str(uuid.uuid4()),
                    time=ttime.time(), owner='arkilic', project='trial',
                    beamline_id='trial_b', updated=False)
    samp = SampleReference(host=TESTING_CONFIG['host'],
                           port=TESTING_CONFIG['port'])
    pytest.raises(HTTPError,
                  samp.update, query={'name': test_sample['name']},
                                      update={'uid': 'illegal'})


def setup():
    amostra_setup()
    global sample_uids, document_insertion_time
    document_insertion_time = ttime.time()
    # TODO: Get sample data!!!!!! and populate



__author__ = 'arkilic'
import setuptools
import versioneer
import os


here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'requirements.txt')) as requirements_file:
    # Parse requirements.txt, ignoring any commented-out lines.
    requirements = [line for line in requirements_file.read().splitlines()
                    if not line.startswith('#')]

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setuptools.setup(
    name='amostra',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license="BSD 3-Clause",
    url="https://github.com/NSLS-II/amostra.git",
    packages=setuptools.find_packages(),
    install_package_data = True,
    install_requires=requirements,
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Development Status :: 3 - Alpha",
        'Programming Language :: Python :: 3',
    ],
)



from pymongo import MongoClient
import time as ttime
from amostra.ignition import start_server
import uuid
import os
import shutil

TESTING_CONFIG = {
    'database': "mds_testing_disposable_{}".format(str(uuid.uuid4())),
    'mongo_server': 'localhost',
    'mongo_port': 27017,
    'host': 'localhost',
    'port': 7770,
    'timezone': 'US/Eastern',
    'mongo_user': 'tom',
    'mongo_pwd': 'jerry',
    'local_files': '~/amostra_files'}


def amostra_setup():
    # start_server(config=TESTING_CONFIG)
    # ensure tornado server started prior to tests
    ttime.sleep(1)


def amostra_teardown():
    uri = 'mongodb://{0}:{1}@{2}:{3}/'.format(TESTING_CONFIG['mongo_user'],
                                              TESTING_CONFIG['mongo_pwd'],
                                              TESTING_CONFIG['mongo_server'],
                                              TESTING_CONFIG['mongo_port'])
    conn = MongoClient(uri)
    conn.amostra.drop_collection('sample')
    conn.amostra.drop_collection('request')
    conn.amostra.drop_collection('container')


def amostra_local_setup():
    try:
        usr_path = os.path.expanduser(TESTING_CONFIG['local_files'])
        os.mkdir(usr_path)
    except FileExistsError:
        pass


def amostra_local_teardown():
    try:
        shutil.rmtree(TESTING_CONFIG['local_files'])
    except FileNotFoundError:
        pass


class _baseSM:
    @classmethod
    def test_create(self):
        db = self.db
        ast_uid = str(uuid.uuid4())
        uid = db.create(name='obelix', location='gaul', occupation='hero',
                        uid=ast_uid)



import time as ttime
import uuid
import pytest
from amostra.testing import amostra_local_setup, amostra_local_teardown
from amostra.client.local_commands import (LocalSampleReference,
                                           LocalContainerReference,
                                           LocalRequestReference)


def setup():
    amostra_local_setup()


def test_constructors():
    # attempt empty reference create
    s_ref = LocalSampleReference()
    c_ref = LocalContainerReference()
    r_ref = LocalRequestReference()


def test_create_sample():
    s = LocalSampleReference()
    s.create(uid=str(uuid.uuid4()), name='local roman',
             time=ttime.time(), compound='Fe', material='sword')


def test_create_request():
    r = LocalRequestReference()
    r.create(sample=None, time=ttime.time(), uid=str(uuid.uuid4()))


def test_create_container():
    c = LocalContainerReference()
    c.create(time=ttime.time(), uid=str(uuid.uuid4()), name='village',
             type='ancient gaul')


def test_find_sample():
    s = LocalSampleReference()
    samp_dict = dict(uid=str(uuid.uuid4()), time=ttime.time(), name='hidefix',
                     kind='dog', breed='multigree')
    s.create(**samp_dict)
    assert next(s.find(uid=samp_dict['uid']))['uid'] == samp_dict['uid']


def test_find_container():
    c = LocalContainerReference()
    cont_dict = dict(uid=str(uuid.uuid4()), time=ttime.time(), name='village',
                     kind='gaul', population='50')
    c.create(**cont_dict)
    assert next(c.find(uid=cont_dict['uid']))['uid'] == cont_dict['uid']


def test_find_request():
    r = LocalRequestReference()
    req_dict = dict(uid=str(uuid.uuid4()), time=ttime.time(), name='war',
                    kind='street fight', state='inactive', winner='gauls')
    r.create(**req_dict)
    assert next(r.find(uid=req_dict['uid']))['uid'] == req_dict['uid']


def test_update_sample():
    s = LocalSampleReference()
    samp = dict(uid=str(uuid.uuid4()), name='julius',
                time=ttime.time(), position='emperor', material='wisdom',
                state='in office')
    s.create(**samp)
    s.update({'uid': samp['uid']},
             {'state': 'murdered by brutus'})
    assert next(s.find(state='murdered by brutus'))['uid'] == samp['uid']


def test_update_container():
    c = LocalContainerReference()
    cont = dict(uid=str(uuid.uuid4()), container=str(uuid.uuid4()),
                time=ttime.time(),state='empty')

    c.create(**cont)
    c.update({'uid': cont['uid']},
             {'state': 'full'})
    assert next(c.find(uid=cont['uid']))['state'] == 'full'


def test_update_request():
    r = LocalRequestReference()
    req = dict(uid=str(uuid.uuid4()), sample=str(uuid.uuid4()),
                time=ttime.time(), scan='mesh')

    r.create(**req)
    r.update({'uid': req['uid']},
             {'scan': 'energy'})
    assert next(r.find(uid=req['uid']))['scan'] == 'energy'


def teardown():
    amostra_local_teardown()
